import re

import weave
from gpt_translate.loader import MDPage


@weave.op
def validate_links(original_page: MDPage, translated_page: MDPage):
    """
    Validate that the links in the original page are the same as the links in the translated page.
    """
    original_links = original_page.links
    translated_links = translated_page.links

    missing_links = [link for link in original_links if link not in translated_links]
    extra_links = [link for link in translated_links if link not in original_links]
    return {
        "links_match": len(missing_links) == 0 and len(extra_links) == 0,
        "missing_links": [l.target for l in missing_links],
        "extra_links": [l.target for l in extra_links],
        "total_links": len(original_links),
    }


@weave.op
def validate_headers(original_page: MDPage, translated_page: MDPage):
    """
    Validate that the headers in the original page are the same as the headers in the translated page.
    """
    original_header = original_page.header
    translated_header = translated_page.header
    title_match = original_header.title == translated_header.title
    description_match = original_header.description == translated_header.description
    slug_match = original_header.slug == translated_header.slug
    displayed_sidebar_match = (
        original_header.displayed_sidebar == translated_header.displayed_sidebar
    )
    imports_match = original_header.imports == translated_header.imports
    return {
        "title_match": title_match,
        "description_match": description_match,
        "slug_match": slug_match,
        "displayed_sidebar_match": displayed_sidebar_match,
        "imports_match": imports_match,
    }


def _validate_tabs_format(content: str) -> bool:
    tab_pattern = re.compile(
        r"""
        <Tabs\s*[^>]*>\s*
        (?:<TabItem\s*[^>]*>\s*.*?\s*</TabItem>\s*)+
        </Tabs>
        """,
        re.DOTALL | re.VERBOSE,
    )
    return bool(tab_pattern.search(content))


@weave.op
def validate_tabs(translated_page: MDPage):
    """
    Validate the Tabs in the docosaurus format
    """
    content = translated_page.content

    return {"tabs_format_valid": _validate_tabs_format(content)}



from gpt_translate.validate import _validate_tabs_format


def test_validate_tabs_format_valid():
    content = """
    <Tabs>
        <TabItem value="python" label="Python">
            Python content
        </TabItem>
        <TabItem value="javascript" label="JavaScript">
            JavaScript content
        </TabItem>
    </Tabs>
    """
    assert _validate_tabs_format(content) == True


def test_validate_tabs_format_invalid():
    content = """
    <Tabs>
        <TabItem value="python" label="Python">
            Python content
        </TabItem>
        <TabItem value="javascript" label="JavaScript">
            JavaScript content
    </Tabs>
    """
    assert _validate_tabs_format(content) == False


def test_validate_tabs_format_with_random_text():
    content = """
    Some random text before the tabs.
    <Tabs>
        <TabItem value="python" label="Python">
            Python content
        </TabItem>
        <TabItem value="javascript" label="JavaScript">
            JavaScript content
        </TabItem>
    </Tabs>
    Some random text after the tabs.
    """
    assert _validate_tabs_format(content) == True



import yaml
import json
import logging
import asyncio
from typing import Optional
from pathlib import Path
from dataclasses import dataclass
from openai import AsyncOpenAI
from tqdm.asyncio import tqdm
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,  # for exponential backoff
)

import weave
from pydantic import model_validator, Field

from gpt_translate.prompts import PromptTemplate
from gpt_translate.loader import (
    remove_markdown_comments,
    MDPage,
    Header,
    split_markdown,
)
from gpt_translate.utils import file_is_empty, count_tokens, remove_after
from gpt_translate.validate import validate_links, validate_headers, validate_tabs


# Use the OpenAI API in async mode
client = AsyncOpenAI()

## Globals
REPLACE = False
REMOVE_COMMENTS = True
MAX_OPENAI_CONCURRENT_CALLS = 7  # Adjust the limit as needed


@retry(wait=wait_random_exponential(min=1, max=60), stop=stop_after_attempt(6))
async def completion_with_backoff(**kwargs):
    return await client.chat.completions.create(**kwargs)


@weave.op
async def longer_create(messages=None, max_tokens=4096, **kwargs):
    """
    longer_create is a function that extends the max_tokens beyond the default 4096 by recursively calling the create method if the finish_reason is hitting the max_tokens.
    """
    if messages is None:
        messages = []

    res = await completion_with_backoff(
        messages=messages, max_tokens=max_tokens, **kwargs
    )
    message_content = res.choices[0].message.content
    logging.debug(res.usage)
    logging.debug(
        f"[blue]OpenAI response:\n{message_content[:100]}...[/blue]",
        extra={"markup": True},
    )

    finish_reason = res.choices[0].finish_reason
    if finish_reason == "length":
        # trim message to the last separator
        process_tail = remove_after(message_content)
        messages.append({"role": "assistant", "content": process_tail["text"]})
        # Recursively call the function with the last assistant's message
        logging.debug(f"Recursively calling with {messages[-1]['content'][:100]}")
        next_response = await longer_create(
            messages=messages, max_tokens=max_tokens, **kwargs
        )
        return process_tail["text"] + next_response
    else:
        return message_content


@dataclass
class TranslationResult:
    content: str
    tokens: int


@weave.op
async def translate_content(
    md_content: str, prompt: PromptTemplate, token_count: int = None, **model_args
):
    """Translate a markdown chunk asynchronously
    md_content: markdown content
    prompt: PromptTemplate object
    return: translated page
    """
    output = await longer_create(
        messages=prompt.format(md_chunk=md_content), **model_args
    )
    return TranslationResult(content=output, tokens=count_tokens(output))


class Translator(weave.Object):
    "A class to translate markdown files asynchronously"
    config_folder: Path
    language: str = "ja"
    do_evaluation: bool = True
    do_translate_header_description: bool = True
    model_args: dict = dict(model="gpt-4o", temperature=1.0)
    evaluation_prompt: Optional[str] = Field(default=None)
    prompt_template: PromptTemplate = Field(default=None)

    @model_validator(mode="before")
    def initialize_fields(cls, values):
        config_folder = Path(values.get("config_folder"))
        language = values.get("language", "ja")
        prompt_template = PromptTemplate.from_files(
            config_folder / "system_prompt.txt",
            config_folder / "human_prompt.txt",
            config_folder / f"language_dicts/{language}.yaml",
        )
        if (config_folder / "evaluation_prompt.txt").exists():
            evaluation_prompt = (config_folder / "evaluation_prompt.txt").read_text()
        else:
            evaluation_prompt = None

        values.update(
            {
                "config_folder": config_folder,
                "prompt_template": prompt_template,
                "evaluation_prompt": evaluation_prompt,
            }
        )
        return values

    # @weave.op
    # async def translate_splitted_md(
    #     self,
    #     splitted_markdown: list[str],
    #     sep: str = "\n\n",
    # ) -> str:
    #     """Translate a list of markdown chunks asynchronously
    #     splitted_markdown: list of markdown chunks
    #     prompt: PromptTemplate object
    #     max_chunk_tokens: maximum number of tokens per chunk
    #     sep: separator between chunks
    #     model_args: arguments to pass to the completion_with_backoff function
    #     return: translated markdown file
    #     """

    #     tasks = []
    #     packed_chunks = ""
    #     packed_chunks_len = 0

    #     translated_chunks = []
    #     for i, chunk in enumerate(splitted_markdown):
    #         n_tokens = count_tokens(chunk)

    #         if packed_chunks_len + n_tokens <= self.max_chunk_tokens:
    #             logging.debug(f"Packing chunk {i} with {n_tokens} tokens")
    #             packed_chunks += sep + chunk
    #             packed_chunks_len += n_tokens
    #         else:
    #             logging.debug(f">> Translating {packed_chunks_len} tokens")
    #             translated_chunk = await translate_content(
    #                 packed_chunks, self.prompt_template, token_count=packed_chunks_len, **self.model_args
    #             )
    #             translated_chunks.append(translated_chunk.content)
    #             logging.debug(f">> Translated {translated_chunk.tokens} tokens")
    #             packed_chunks = chunk
    #             packed_chunks_len = n_tokens

    #     if packed_chunks:
    #         logging.debug(f">> Translating {packed_chunks_len} tokens (last chunk)")
    #         translated_chunk = await translate_content(
    #             packed_chunks, self.prompt_template, token_count=packed_chunks_len, **self.model_args
    #         )
    #         translated_chunks.append(translated_chunk.content)
    #     return sep.join(translated_chunks)

    @weave.op
    async def translate_file(self, md_file: str, remove_comments: bool = True):
        """Translate a markdown file asynchronously"""
        with open(md_file, "r") as f:
            md_content = f.read()
        if remove_comments:
            logging.debug("Removing comments")
            md_content = remove_markdown_comments(md_content)
        if len(md_content.strip()) < 20:
            logging.warning(f"File may be empty: {md_file}")
        md_page = MDPage.from_raw_content(filename=md_file, raw_content=md_content)
        logging.debug(
            f"[bold red blink]Calling OpenAI [/bold red blink]with {self.model_args}\nFile: {md_file}\nContent:\n{md_content[:100]}...",
            extra={"markup": True},
        )
        translated_page = await self.translate_page(md_page)

        if self.evaluate:
            evaluation_results = await self.evaluate(md_page, translated_page)
            return {
                "translated_page": translated_page,
                "evaluation_results": evaluation_results,
            }
        return {"translated_page": translated_page}

    @weave.op
    async def translate_page(self, md_page: MDPage, translate_header: bool = True):
        """Translate a markdown page asynchronously"""
        # chunks = split_markdown(md_page.content)
        # translated_content = await self.translate_splitted_md(chunks)
        translated_content = await translate_content(
            md_page.content, self.prompt_template, **self.model_args
        )
        logging.debug(f"Translated content: {translated_content}")
        if (
            md_page.header.description and self.do_translate_header_description
        ):  # check if header contains a description
            translated_header_description = await self.translate_header_description(
                md_page
            )
            logging.debug(
                f"Translating header description: {md_page.header.description}"
            )
            new_header = Header(
                title=md_page.header.title,
                description=(
                    translated_header_description.content
                    if md_page.header.description
                    else None
                ),
                slug=md_page.header.slug,
                displayed_sidebar=md_page.header.displayed_sidebar,
                imports=md_page.header.imports,
            )
        else:
            new_header = md_page.header
        return MDPage(
            filename=md_page.filename,
            content=translated_content.content,
            header=new_header,
        )

    @weave.op
    async def translate_header_description(self, md_page: MDPage):
        """Translate the header description"""
        translated_description = await translate_content(
            md_page.header.description, self.prompt_template, **self.model_args
        )
        return translated_description

    @weave.op
    async def evaluate(self, md_page: MDPage, translated_page: MDPage):
        """Validate the translation"""
        links_validation = validate_links(md_page, translated_page)
        headers_validation = validate_headers(md_page, translated_page)
        tabs_validation = validate_tabs(translated_page)
        logging.debug(f"✅ Links validation: {links_validation}")
        logging.debug(f"✅ Headers validation: {headers_validation}")
        logging.debug(f"✅ Tabs validation: {tabs_validation}")
        translation_validation = await self.evaluate_translation(
            md_page, translated_page
        )
        logging.debug(f"✅ Translation validation: {translation_validation}")
        return {
            "links_validation": links_validation,
            "headers_validation": headers_validation,
            "tabs_validation": tabs_validation,
            "translation_validation": translation_validation,
        }

    @weave.op
    async def evaluate_translation(self, md_page: MDPage, translated_page: MDPage):
        """Evaluate the translation"""
        messages = self.prompt_template.format(md_chunk=md_page.content)
        messages.append({"role": "assistant", "content": f"{translated_page.content}"})
        messages.append({"role": "user", "content": self.evaluation_prompt})
        res = await completion_with_backoff(
            messages=messages,
            **self.model_args,
            response_format={"type": "json_object"},
        )
        extracted = res.choices[0].message.content
        analysis = json.loads(extracted)
        return analysis


@weave.op
async def _translate_file(
    input_file: str,  # File to translate
    out_file: str,  # File to save the translated file to
    replace: bool = REPLACE,  # Replace existing file
    language: str = "es",  # Language to translate to
    config_folder: str = "./configs",  # Config folder
    remove_comments: bool = REMOVE_COMMENTS,  # Remove comments
    do_evaluation: bool = True,  # Evaluate the translated file
    do_translate_header_description: bool = True,  # Translate the header description
    model_args: dict = dict(model="gpt-4o", temperature=1.0),  # model args
) -> MDPage:
    """Translate a markdown file asynchronously"""
    if file_is_empty(input_file):
        raise ValueError(f"File {input_file} is empty")

    # check it is a md file
    if Path(input_file).suffix != ".md":
        raise ValueError(f"File {input_file} is not a markdown file")
    out_file = Path(out_file)
    if out_file.exists() and not replace and not file_is_empty(out_file):
        logging.info(f"File {out_file} already exists. Use --replace to overwrite.")
    else:
        out_file.parent.mkdir(parents=True, exist_ok=True)
        try:
            translator = Translator(
                config_folder=config_folder,
                language=language,
                do_evaluation=do_evaluation,
                do_translate_header_description=do_translate_header_description,
                model_args=model_args,
            )
            translation_results = await translator.translate_file(
                input_file, remove_comments
            )
            with open(out_file, "w", encoding="utf-8") as f:
                f.write(str(translation_results["translated_page"]))
            logging.info(
                f"✅ Translated file saved to [green]{out_file}[/green]",
                extra={"markup": True},
            )
            return translation_results
        except Exception as e:
            logging.error(f"❌ Error translating {input_file}: {e}")
            raise e


@weave.op
async def _translate_files(
    input_files: list[str],  # Files to translate
    input_folder: str,  # folder where the file lives
    out_folder: str,  # Folder to save the translated files to
    replace: bool = REPLACE,  # Replace existing file
    language: str = "es",  # Language to translate to
    config_folder: str = "./configs",  # Config folder
    remove_comments: bool = REMOVE_COMMENTS,  # Remove comments
    do_evaluation: bool = True,  # Evaluate the translated file
    do_translate_header_description: bool = True,  # Translate the header description
    model_args: dict = dict(model="gpt-4o", temperature=1.0),  # model args
    max_openai_concurrent_calls: int = MAX_OPENAI_CONCURRENT_CALLS,  # Maximum number of concurrent calls to OpenAI
):
    # let's make input_files support a txt file with a list of files
    if not isinstance(input_files, list):
        if Path(input_files).suffix == ".txt":
            logging.info(f"Reading {input_files}")
        input_files = Path(input_files).read_text().splitlines()
    input_files = [
        Path(f) for f in input_files if (Path(f).suffix == ".md" and Path(f).exists())
    ]
    logging.info(
        f"Translating {len(input_files)} file" + ("s" if len(input_files) > 1 else "")
    )
    input_files.sort()
    input_folder = Path(input_folder)
    out_folder = Path(out_folder)
    if not input_folder.is_dir():
        raise ValueError(f"{input_folder} is not a folder")

    semaphore = asyncio.Semaphore(max_openai_concurrent_calls)

    async def _translate_with_semaphore(md_file):
        async with semaphore:
            out_file = out_folder / md_file.relative_to(input_folder)
            return await _translate_file(
                input_file=str(md_file),
                out_file=str(out_file),
                replace=replace,
                language=language,
                config_folder=config_folder,
                remove_comments=remove_comments,
                do_evaluation=do_evaluation,
                do_translate_header_description=do_translate_header_description,
                model_args=model_args,
            )

    tasks = [_translate_with_semaphore(md_file) for md_file in input_files]

    await tqdm.gather(*tasks, desc="Translating files")


if __name__ == "__main__":
    from gpt_translate.cli import setup_logging

    setup_logging(debug=True, silence_openai=True, weave_project="gpt-translate")
    asyncio.run(
        _translate_files(
            input_files="one.txt",
            input_folder="../docodile/docs_main/",
            out_folder="../docodile/docs/",
            replace=True,
            language="ja",
            config_folder="./configs",
            do_evaluation=True,
        )
    )



# This file is in sync with the config.yaml file

import shlex
import logging
from dataclasses import dataclass
import simple_parsing
from simple_parsing.helpers import Serializable

DEFAULT_CONFIG_PATH = "./configs/config.yaml"


@dataclass
class Config(Serializable):
    model: str  # Model to use
    temperature: float  # Temperature to use
    max_tokens: int  # Max tokens to use

    debug: bool  # Debug mode
    weave_project: str  # Weave project
    silence_openai: bool  # Silence OpenAI logger

    language: str  # Language to translate to
    config_folder: str  # Config folder
    replace: bool  # Replace existing file
    remove_comments: bool  # Remove comments
    do_evaluation: bool  # Do evaluation
    do_translate_header_description: bool  # Translate the header description
    max_openai_concurrent_calls: int  # Max number of concurrent calls to OpenAI

    input_file: str  # File to translate, can be a .txt file with a list of files when used with translate.files
    out_file: str  # File to save the translated file to
    input_folder: str  # Folder to translate
    out_folder: str  # Folder to save the translated files to
    limit: int = None  # Limit number of files to translate


def setup_parsing(args):
    # parser = simple_parsing.ArgumentParser(
    #     config_path=DEFAULT_CONFIG_PATH, add_config_path_arg=True
    # )
    # parser.add_arguments(LogsConfig, dest="logs")
    # parser.add_arguments(TranslationConfig, dest="translation")
    # parser.add_arguments(FilesConfig, dest="files")
    # parser.add_arguments(ModelConfig, dest="model")

    config = simple_parsing.parse(
        config_path=DEFAULT_CONFIG_PATH,
        config_class=Config,
        args=args,
        add_config_path_arg=True,
    )
    # logs_args: LogsConfig = args.logs
    # translation_args: TranslationConfig = args.translation
    # file_args: FilesConfig = args.files
    # model_args: ModelConfig = args.model
    # return logs_args, translation_args, file_args, model_args
    return config



import re, yaml
import logging
from typing import Optional
from dataclasses import dataclass, asdict

import weave
from pydantic import model_validator, Field


def remove_markdown_comments(content):
    # Pattern to match HTML comment blocks
    comment_pattern = re.compile(r"<!--.*?-->", re.DOTALL)

    # Remove all matched comment blocks
    cleaned_content = re.sub(comment_pattern, "\n", content)

    return cleaned_content


def split_markdown(content):
    header_pattern = re.compile(r"^(#{1,6} .+)$", re.MULTILINE)
    code_block_pattern = re.compile(r"^```")

    chunks = []
    current_chunk = []
    in_code_block = False

    for line in content.split("\n"):
        if code_block_pattern.match(line):
            in_code_block = not in_code_block

        # Split at headers, only if not within a code block
        if header_pattern.match(line) and not in_code_block:
            if current_chunk:
                # Add the current chunk to chunks
                chunks.append("\n".join(current_chunk).strip())
                current_chunk = [line]
                in_code_block = False
            else:
                current_chunk.append(line)
        else:
            current_chunk.append(line)

    # Add the last chunk if not empty
    if current_chunk:
        chunks.append("\n".join(current_chunk).strip())

    return chunks


@dataclass
class MDLink:
    title: str
    target: str
    filename: str
    line_number: str

    def __str__(self):
        return f"{self.filename}:{self.line_number:>4}: [{self.title}]({self.target})"

    def __eq__(self, other):
        "We only care to know if the link is still pointing to the same place"
        return self.target == other.target


def extract_markdown_links(filename, content):
    # This regular expression matches the Markdown link syntax
    link_pattern = r"\[([^\]]+)\]\(([^)]+)\)"
    links = []
    for i, line in enumerate(content.split("\n")):
        matches = re.findall(link_pattern, line)
        for title, target in matches:
            links.append(MDLink(title, target, filename, i + 1))
    return links


@dataclass
class Header:
    title: Optional[str] = None
    description: Optional[str] = None
    slug: Optional[str] = None
    displayed_sidebar: Optional[str] = None
    imports: Optional[str] = None

    @classmethod
    def from_string(cls, input_string: str):
        # Split the string into lines
        lines = input_string.splitlines()

        # Identify and separate the import lines and the YAML content
        import_lines = []
        yaml_lines = []
        yaml_start = False
        for line in lines:
            if line.strip() == "---":
                yaml_start = not yaml_start
                continue
            if yaml_start:
                yaml_lines.append(line)
            else:
                import_lines.append(line)

        # Parse the YAML content
        attributes = yaml.safe_load("\n".join(yaml_lines)) or {}
        # Rejoin the import lines into a single string
        imports = "\n".join(import_lines).strip()
        if not attributes and not imports:
            return cls()
        return cls(**attributes, imports=imports)

    def __repr__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        parts = []
        attrs = {k: v for k, v in asdict(self).items() if v and k != "imports"}
        # sort by key
        if attrs:
            attr_parts = ""
            attr_parts += "---\n"
            for key, value in attrs.items():
                # let's handle if value is a multiline string by removing the line breaks
                if isinstance(value, str):
                    value = value.strip().replace("\n", " ")
                attr_parts += f"{key}: {value}\n"
            attr_parts += "---"
            parts.append(attr_parts)
        if self.imports:
            parts.append(self.imports)
        header = "\n".join(parts).encode("utf-8").decode("utf-8")
        return header


@weave.op
def extract_header(content: str) -> dict:
    "Extract header from a markdown file, including YAML frontmatter and imports"
    lines = content.split("\n")
    frontmatter = []
    imports = []
    content_lines = []
    in_frontmatter = False

    # Extract frontmatter
    for line in lines:
        if line.strip() == "---":
            if not in_frontmatter:
                in_frontmatter = True
                frontmatter.append(line)
            else:
                frontmatter.append(line)
                break
        elif in_frontmatter:
            frontmatter.append(line)
        else:
            break

    # Extract imports and content
    remaining_lines = lines[len(frontmatter) :]
    for line in remaining_lines:
        if line.strip().startswith("import "):
            imports.append(line)
        else:
            content_lines = remaining_lines[len(imports) :]
            break

    # Combine frontmatter and imports for header
    header_lines = frontmatter + imports

    # Remove trailing empty lines from header
    while header_lines and not header_lines[-1].strip():
        header_lines.pop()

    header = "\n".join(header_lines).rstrip()
    content = "\n".join(content_lines).strip()

    return {"header": header, "content": content}


@weave.op
def find_links(raw_content: str, filename: str) -> list[MDLink]:
    """
    Finds all Markdown links in the content.
    :return: list of tuples, each containing (link text, URL).
    """
    link_pattern = r"\[([^\]]+)\]\(([^)]+)\)"
    links = []
    for i, line in enumerate(raw_content.split("\n")):
        matches = re.findall(link_pattern, line)
        for title, target in matches:
            links.append(MDLink(title, target, filename, i + 1))
    return links


class MDPage(weave.Object):
    filename: str
    content: str
    header: Header
    links: list[MDLink] = Field(default=None)

    @model_validator(mode="after")
    def set_links(self):
        self.links = find_links(self.content, self.filename)
        return self

    @classmethod
    def from_raw_content(cls, filename: str, raw_content: str) -> "MDPage":
        extracted = extract_header(raw_content)
        header = extracted["header"]
        content = extracted["content"]
        header = Header.from_string(header)
        return cls(
            filename=filename,
            content=content,
            header=header,
        )

    @weave.op
    def update_links(self, new_links: list[MDLink], targets_only=True) -> None:
        "Update the links in the content"
        if len(new_links) == len(self.links):
            logging.error(
                f"The following links don't match: {new_links} vs {self.links}"
            )
            raise ValueError(
                f"Number of links don't match: {len(new_links)} vs {len(self.links)}"
            )
        logging.debug(f"Maybe updating links in {self.filename}")
        for old_link, new_link in zip(self.links, new_links):
            if old_link.target != new_link.target:
                logging.debug(f"Replacing {old_link} with {new_link}")
                self.content = self.content.replace(old_link.target, new_link.target)
        if targets_only:
            self.links = self.find_links(self.content)
        else:
            self.links = new_links

    def __str__(self):
        "Concatenate header and content"
        if str(self.header).strip():
            return f"{self.header}\n\n{self.content}"
        return str(self.content)
